#-*- coding:utf-8 -*-

#  Copyright (C) 2009 B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filename: flubrd.scm
# Original copyright and license:
#;;   Copyright (c) 2004 Dale Mellor
#;;
#;;   This program is free software; you can redistribute it and/or modify
#;;   it under the terms of the GNU General Public License as published by
#;;   the Free Software Foundation; either version 3 of the License, or
#;;   (at your option) any later version.
#;;
#;;   This program is distributed in the hope that it will be useful,
#;;   but WITHOUT ANY WARRANTY; without even the implied warranty of
#;;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#;;   GNU General Public License for more details.
#;;
#;;   You should have received a copy of the GNU General Public License
#;;   along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Direct access to a cube object. It is assumed the user knows what he is doing
# (read the manual!) :-)

def get_colour (cube, face, index):
    #print 'cube',cube
    return cube.faces[face][index]

def set_colour_p (cube, face, index, colour):
    #print 'cube',cube
    cube.faces[face][index] = colour


# All the rest of this file assumes a certain structure to a cube object. This
# test makes sure that structure is in place. The cube structure passed from
# the C world starts with a list of structure version, number of dimensions,
# and the size of the cube in each dimension.
#
# This function is not applied implicitly for efficiency's sake, but should be
# exported from the module (when it becomes one) so that the user can check the
# applicability of the module).
def check_cube_structure (cube):
    #print cube
    if cube.dimension == 3:
        return True
    else:
        plugin_error_dialog(_('This script only works on 3x3x3 cubes.'))
        return False


# This list defines the relationships and orientations of the faces.
# The key is the number of
# the face, and then the values are the letter used to symbolically identify
# the face, the face which runs along the 'top' edge of this face (the row
# holding the 0, 1 and 2 cells), and the face which runs along the 'right' edge
# of this face (the column with the 2, 5 and 8 cells).
#
# These correspondences have been determined by trial and error. There is also
# an implicit assumption made that faces 0 and 1, 2 and 3, and faces 4 and 5
# are opposites.

faces = ((4,  ('f', 0, 3)),
                (5,  ('b', 0, 3)),
                (2,  ('l', 0, 5)),
                (3,  ('r', 0, 5)),
                (0,  ('u', 2, 5)),
                (1,  ('d', 2, 5)))


# Return the number of the face opposite the one given.
def opposite_face (face):
    return (face // 2) * 2 + 1 - face % 2


# Alist with fbdurl strings to identify elements as the key, and a cons cell of
# face and index as the value.

symbolic_mapping = []


# A convenience to help construct the above alist.
def _add_map (symbol, face, index):
    global symbolic_mapping
    #print '79', symbol, face, index
    symbolic_mapping.insert(0, (symbol, (face, index)))
add_map = _add_map


# For every face, make a mapping for each cell.
def _proc():
    def lambda_(face):
        #print '86',face
        #print faces
        #global add_map
        add_map(face[1][0], face[0], 4)
        add_map_ = lambda direction, opposite, index: (
                    add_map(
                        face[1][0] + dict(faces)[
                                           (opposite_face if opposite else (lambda x: x))(
                                                        (face[1][2] if direction else face[1][1]))
                                        ][0],
                        face[0],
                        index
                    )
                )
        add_map_(False, False, 1)
        add_map_(True, False, 5)
        add_map_(False, True, 7)
        add_map_(True, True, 3)
        o = lambda x: (opposite_face if x else lambda x: x)
        add_map__ = lambda opp1, opp2, index: (
                            add_map(
                                face[1][0] +
                                dict(faces)[ o(opp1)(face[1][1]) ][0] +
                                dict(faces)[ o(opp2)(face[1][2]) ][0],
                                face[0],
                                index)
                        )
        add_map__(False, False, 2)
        add_map__(True, False, 8)
        add_map__(False, True, 0)
        add_map__(True, True, 6)
        o = lambda x: (opposite_face if x else lambda x: x)
        add_map___ = lambda opp1, opp2, index: (add_map(
                                face[1][0] +
                                dict(faces)[ o(opp2)(face[1][2]) ][0] +
                                dict(faces)[ o(opp1)(face[1][1]) ][0],
                                face[0],
                                index
                            ))
        add_map___(False, False, 2)
        add_map___(True, False, 8)
        add_map___(False, True, 0)
        add_map___(True, True, 6)
    return lambda_
for _x in (faces):
    _proc()(_x)


# This is what it's all about. These are the two main functions which should be
# exported from this module (when we make it a module).
def get_colour_symbolic (cube, symbol):
    #print 'symbol',symbol
    #print 'symbolic_mapping',symbolic_mapping
    symbol_entry = symbol, dict(symbolic_mapping)[symbol]
    #print 'symbol_entry',symbol_entry
    return get_colour(cube, symbol_entry[1][0], symbol_entry[1][1])

def set_colour_symbolic_p (cube, symbol, colour):
    symbol_entry = symbol, dict(symbolic_mapping)[symbol]
    set_colour_p(cube, symbol_entry[1][0], symbol_entry[1][1], colour)


#----------------------------------------------------------------------
#
# The second part of this file defines the functions needed to move the cube
# about.
#
#----------------------------------------------------------------------


# Create a procedure which rotates the letters of a string according to the
# specification passed: l is a list of pairs of letters with the notion that
# occurrences of the first should be replaced with the second, in the strings
# that the returned procedure processes.
def define_face_twister (sub_list):
    #print 'sub_list',sub_list
    def lambda_((s)):
        res = ''
        s = s[:]
        chars = len(s)
        for i in xrange(chars):
            #print 'sub_list',sub_list, 's',s, i
            replacement = dict(sub_list).get(s[i])
            #print 'replacement',replacement
            if replacement:
                #s = s.string_set(i,replacement)
                res += replacement
            else:
                res += s[i]
        return res
    return lambda_
  

# In its natural state, this procedure manipulates the given cube object
# to reflect an anticlockwise rotation of the front face. However, the supplied
# face-twister function may conspire to fiddle the face letters whenever the
# cube is accessed so that, in face, any face may be the one being rotated.
def twist_front (cube, face_twister):
    #print 'face_twister',face_twister
    def _proc():
        def lambda_(twists):
            #print 'twists',twists
            #print 'face_twister',face_twister
            first = get_colour_symbolic(cube, face_twister(twists[0]))
            def loop(twists=twists):
                if len (twists) == 1:
                    set_colour_symbolic_p(cube, face_twister(twists[0]), first)
                else:
                    #begin
                    set_colour_symbolic_p(cube, face_twister(twists[0]),
                                            get_colour_symbolic(cube, face_twister(twists[1])))
                    loop(twists[1:])
            loop()
        return lambda_
    for _x in ((    ('ufr', 'rfd', 'dfl', 'lfu'),
                    ('uf', 'rf', 'df', 'lf'),
                    ('ufl', 'rfu', 'dfr', 'lfd'),
                    ('ful', 'fur', 'frd', 'fld'),
                    ('fu', 'fr', 'fd', 'fl'))):
        _proc()(_x)


# Wrapper around the procedure above which works out a face-twister function
# given the face the user wants rotating, and works out how many turns are
# required given the user's notion of clockwise quarters.
def twist_face (cube, face, quarters):
    def _func():
        if   face[0] == ('f'): return ()
        elif face[0] == ('b'): return (('b',  'f'), ('f',  'b'), ('l',  'r'), ('r',  'l'))
        elif face[0] == ('u'): return (('u',  'b'), ('b',  'd'), ('d',  'f'), ('f',  'u'))
        elif face[0] == ('d'): return (('d',  'b'), ('b',  'u'), ('u',  'f'), ('f',  'd'))
        elif face[0] == ('l'): return (('l',  'b'), ('b',  'r'), ('r',  'f'), ('f',  'l'))
        elif face[0] == ('r'): return (('r',  'b'), ('b',  'l'), ('l',  'f'), ('f',  'r'))
    face_twister = define_face_twister(_func())
    def _func_():
        if quarters == (1): return 3
        if quarters == (2): return 2
        if quarters == (3): return 1
        if quarters == (-1): return 1
    
    i = _func_()
    while i != 0:
        twist_front(cube, face_twister)
        i = (i-1)


# Implementation of the move buffer object. This list stores the moves
# generated by the move-face procedure for eventual submission to the
# gnubik-rotate-* procedures.

move_buffer = []

def execute_move_buffer_p():
    global move_buffer
    plugin_rotate_animated(list(reversed(move_buffer)))
    move_buffer = []

def clear_move_buffer_p():
    global move_buffer
    move_buffer = []


# A procedure which applies the above function for specific symbolic move
# names. It also returns an entry that should be passed in a list to
# gnubik-rotate-animated et al., to cause the cube to be rotated
# also.
#
# This is the third main function which should be exported from this module
# (when it becomes one).
def move_face (cube, symbol):
    data = symbol, dict([   ('f',   (1, (0, -2, 0))),
                            ('b',   (1, (0,  2, 1))),
                            ('l',   (1, (1, -2, 0))),
                            ('r',   (1, (1,  2, 1))),
                            ('u',   (1, (2, -2, 0))),
                            ('d',   (1, (2,  2, 1))),
                            ('f+',  (1, (0, -2, 0))),
                            ('b+',  (1, (0,  2, 1))),
                            ('l+',  (1, (1, -2, 0))),
                            ('r+',  (1, (1,  2, 1))),
                            ('u+',  (1, (2, -2, 0))),
                            ('d+',  (1, (2,  2, 1))),
                            ('f ',  (1, (0, -2, 0))),
                            ('b ',  (1, (0,  2, 1))),
                            ('l ',  (1, (1, -2, 0))),
                            ('r ',  (1, (1,  2, 1))),
                            ('u ',  (1, (2, -2, 0))),
                            ('d ',  (1, (2,  2, 1))),
                            ('f-',  (3, (0, -2, 1))),
                            ('b-',  (3, (0,  2, 0))),
                            ('l-',  (3, (1, -2, 1))),
                            ('r-',  (3, (1,  2, 0))),
                            ('u-',  (3, (2, -2, 1))),
                            ('d-',  (3, (2,  2, 0)))
                        ])[symbol]
                
    #print 'move_buffer',move_buffer
    #print 'data', data
    twist_face(cube, symbol, data[1][0])
    #move_buffer.insert(0, data[1][1][0])
    move_buffer.insert(0, data[1][1])
    

