#-*- coding:utf-8 -*-

#  Copyright (C) 2009 B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filename: mellor-solve.scm
# Original copyright and license:
#;;   Copyright (c) 2004 Dale Mellor
#;;
#;;   This program is free software; you can redistribute it and/or modify
#;;   it under the terms of the GNU General Public License as published by
#;;   the Free Software Foundation; either version 3 of the License, or
#;;   (at your option) any later version.
#;;
#;;   This program is distributed in the hope that it will be useful,
#;;   but WITHOUT ANY WARRANTY; without even the implied warranty of
#;;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#;;   GNU General Public License for more details.
#;;
#;;   You should have received a copy of the GNU General Public License
#;;   along with this program.  If not, see <http://www.gnu.org/licenses/>.


#import sys,os
#print os.getcwd()
#print sys.path
#print sorted(sys.modules.keys())
from flubrd_e import *


# I'm not normally one for global variables, but this is so pervasive
# throughout this unit I couldn't resist. It is the reflection of
# GNUbik's own cube object.
cube = None


# Whenever we access the cube, or perform moves on it, we call a function which
# performs symbolic rotations about the sides l, r, b and f. This way, we only
# have to work out one quarter of the total moves to solve the cube, and the
# rest come by symmetry. The following variable gets a function which performs
# the rotations (it will be a function which takes a string in and gives a
# transformed string back).
rotated_flubrd_symbol = None


# Function to create the procedure that gets assigned to the variable binding
# above, given the number of turns required. For example, if one turn is
# requested, then any algorithms which are centred around the front face will,
# unwittingly, be applying themselves to the right face.
def rotated_flubrd_symbol_set_p (turns):
    def rotated_flubrd_symbol_ (s):
        for j in xrange(turns):
            res = ''
            for c in s:
                res += {'l': 'f', 'f': 'r', 'r': 'b', 'b': 'l'}.get(c,c)
            s = res
        return s
    global rotated_flubrd_symbol
    rotated_flubrd_symbol = rotated_flubrd_symbol_


# So, if proc is designed to solve a block on the front face, we apply it to
# four different rotations of the symbolic letters and have it solve a block on
# four faces.
def repeat_for_all_rotations (proc):
    for i in xrange(4):
        rotated_flubrd_symbol_set_p(i)
        proc()


# Wrapper around get-colour-symbolic which applies the cube rotation first,
# i.e. it is influenced by the script later setting rotated-flubrd-symbol by
# calling rotated-flubrd-symbol-set!
def lookup_colour (s):
    return get_colour_symbolic(cube, rotated_flubrd_symbol(s))


# Wrapper around move-face which applies the cube rotation first. Note that
# this takes a string consisting of multiple moves, each one represented by two
# letters, e.g. "f+"; the single-letter symbols which move-face accepts cannot
# be used here.
def do_moves (moves):
    while moves:
        move_face(cube, rotated_flubrd_symbol(moves[0:2]))
        moves = moves[2:]


# This introduces xsubstring, which allows us to perform cyclic rotations on
# the characters of a string.
#use_modules(srfi, srfi_13)


#hink about fixing the uf block, but then apply the algorithms with the cube
#ymbolically rotated four times. Look for the required block in all possible
#ocations, and when it is found lookup the moves that are needed to bring it
# to uf, without upsetting any of the other top edges.
def mellor_top_edge_solve ():
    #print 'mellor_top_edge_solve'
    def _func():
        u = lookup_colour('u')
        f = lookup_colour('f')
        moves = [
                ('fu', 'f-u+l-u-'),
                
                ('ur', 'r-u-r+u+'),
                ('ru', 'r-f-'),
                ('bu', 'b-u-r-u+'),
                ('ub', 'b-u-u-b+u-u-'),
                ('ul', 'l+u+l-u-'),
                ('lu', 'l+f+'),
                
                ('fr', 'u-r+u+'),
                ('rf', 'f-'),
                ('rb', 'r+r+f-r-r-'),
                ('br', 'u-r-u+'),
                ('lb', 'l+l+f+l-l-'),
                ('bl', 'u+l+u-'),
                ('fl', 'u+l-u-'),
                ('lf', 'f+'),
                
                ('fd', 'd+r+f-r-'),
                ('df', 'f+f+'),
                ('rd', 'r+f-r-'),
                ('dr', 'd-f+f+'),
                ('db', 'd+d+f+f+'),
                ('bd', 'd-r+f-r-'),
                ('dl', 'd+f+f+'),
                ('ld', 'l-f+l+'),
            ]
        #def trial (moves=moves):
        #    #print 'trial', str(moves)[:70]
        #    if moves:
        #        #print 'u',u,'->',lookup_colour(moves[0][0]), 'f',f,'->',lookup_colour((moves[0][0]*2)[1:3])
        #        if (u == lookup_colour(moves[0][0]) and
        #           f == lookup_colour((moves[0][0]*2)[1:3])):
        #            print 'do_moves', moves[0][1]
        #            do_moves(moves[0][1])
        #        else:
        #            trial(moves[1:])
        #trial()
        for move in moves:
            if (u == lookup_colour(move[0]) and
               f == lookup_colour((move[0]*2)[1:3])):
                #print 'do_moves', move[1]
                do_moves(move[1])
                break
            
    repeat_for_all_rotations(_func)


#Concentrate on getting the block ufr correct, but then apply the procedure to
#all four sides of the cube. Look for the correct block in all locations, and
# then lookup the moves required to get it to ufr.
def mellor_top_corner_solve ():
    def _func():
        top_colour = lookup_colour('u')
        front_colour = lookup_colour('f')
        right_colour = lookup_colour('r')
        moves = (  ('rfd',    'r-d-r+'),
                   ('fld',  'd+r-d-r+'),
                   ('lbd',  'f+d-d-f-'),
                   ('brd',  'f+d-f-'),
                   
                   ('fdr',    'f+d+f-'),
                   ('rdb',  'd-f+d+f-'),
                   ('bdl',  'r-d-d-r+'),
                   ('ldf',  'r-d+r+'),
                   
                   ('drf',      'r-d+r+f+d-d-f-'),
                   ('dfl',    'd+r-d+r+f+d-d-f-'),
                   ('dlb',  'd-d-r-d+r+f+d-d-f-'),
                   ('dbr',    'd-r-d+r+f+d-d-f-'),

                   ('fru',  'f+d-d-f-r-d-d-r+'),
                   ('ruf',  'r-d-d-r+f+d-d-f-'),
                   ('urb',  'b-d-d-b+r-d+r+'),
                   ('rbu',  'r+d+r-r-d-d-r+'),
                   ('bur',  'f+b-d-f-b+'),
                   ('ubl',  'b+d-b-r-d-d-r+'),
                   ('blu',  'b+r-d-d-r+b-'),
                   ('lub',  'l-f+d-d-f-l+'),
                   ('ulf',  'l+d-l-r-d+r+'),
                   ('lfu',  'l+r-d+r+l-'),
                   ('ful',  'f-d-f+f+d-d-f-'))
        #def loop(moves=moves):
        #    if moves:
        #        if (    lookup_colour(moves[0][0]) == top_colour and 
        #                lookup_colour((moves[0][0]*2)[1:4]) == front_colour and
        #                lookup_colour((moves[0][0]*2)[2:5]) == right_colour
        #            ):
        #            do_moves(moves[0][1])
        #        else:
        #            loop(moves[1:])
        #loop()
        for move in moves:
            if (    lookup_colour(move[0]) == top_colour and 
                    lookup_colour((move[0]*2)[1:4]) == front_colour and
                    lookup_colour((move[0]*2)[2:5]) == right_colour
                ):
                do_moves(move[1])
                break
    repeat_for_all_rotations(_func)


# Consider the fr block. If the block is to be found in the bottom slice, get
# it into one of two `starting positions' (either fd or rd) and then move it up
# into fr by looping back. If the piece is not found here, it must be in one of
# the other middle edge positions; for now we just check if we are holding a
# middle edge block at fr and if so we drop it to the bottom slice, so that a
# repeat of the algorithm will eventually put it in its correct place, and by
# the time the repeat comes around our own block will hopefully have been
# dropped down.
def mellor_middle_slice_solve ():
    for i in xrange(3):
        def _func():
            front_colour = lookup_colour('f')
            right_colour = lookup_colour('r')
            while True:
                if lookup_colour('fd') == front_colour and lookup_colour('df') == right_colour:
                    do_moves('d-r-d+r+d+f+d-f-')
                    break
                elif lookup_colour('df') == front_colour and lookup_colour('fd') == right_colour:
                    do_moves('d+')
                elif lookup_colour('rd') == right_colour and lookup_colour('dr') == front_colour:
                    do_moves('d+f+d-f-d-r-d+r+')
                    break
                elif lookup_colour('dr') == right_colour and lookup_colour('rd') == front_colour:
                    do_moves('d-')
                elif lookup_colour('ld') == front_colour and lookup_colour('dl') == right_colour:
                    do_moves('d+')
                elif lookup_colour('dl') == front_colour and lookup_colour('ld') == right_colour:
                    do_moves('d+d+')
                elif lookup_colour('bd') == front_colour and lookup_colour('db') == right_colour:
                    do_moves('d+d+')
                elif lookup_colour('db') == front_colour and lookup_colour('bd') == right_colour:
                    do_moves('d-')
                
                elif lookup_colour('fr') == front_colour and lookup_colour('rf') == right_colour:
                    break
                else:
                    bottom_colour = lookup_colour('d')
                    if lookup_colour('fr') != bottom_colour and lookup_colour('rf') != bottom_colour:
                        do_moves('d+f+d-f-d-r-d+r+')
                    else:
                        # not found retry in the next round
                        break
                    
        repeat_for_all_rotations(_func)


# Procedure which answers the question, `Is the fdr block located in the right
# place (regardless of orientation)?'
def placed_fdr_q ():
    return (
                    (1 << lookup_colour('fdr')) +
                    (1 << lookup_colour('dfr')) +
                    (1 << lookup_colour('rdf'))
            ) == (
                    (1 << lookup_colour('f')) +
                    (1 << lookup_colour('d')) +
                    (1 << lookup_colour('r'))
                )


# We look at all the bottom corners, and set a bit in a bit-mask when one is in
# the right place. There are six patterns to look for: four in which two
# adjacent blocks are out of place, and two in which diagonal blocks are out of
# place. When we find one of these, we apply a symbolic rotation to the cube so
# that the missing pieces are in set positions, and then apply appriopriate set
# moves. Otherwise, either all the pieces are already in place, or else we just
# shift the bottom slice around and loop until we find a pattern.
def mellor_bottom_corner_place ():
    fix_2 = 'r-d-r+f+d+f-r-d+r+d-d-'
    fix_d = 'r-d-r+f+d-d-f-r-d+r+d-'
    while True:
        a = 0
        for i in xrange(4):
            rotated_flubrd_symbol_set_p(i)
            if placed_fdr_q():
                a += 1 << i
        if   a == 3:  rotated_flubrd_symbol_set_p(3), do_moves(fix_2)
        elif a == 6:  rotated_flubrd_symbol_set_p(0), do_moves(fix_2)
        elif a == 12: rotated_flubrd_symbol_set_p(1), do_moves(fix_2)
        elif a == 9:  rotated_flubrd_symbol_set_p(2), do_moves(fix_2)
        elif a == 5:  rotated_flubrd_symbol_set_p(1), do_moves(fix_d)
        elif a == 10: rotated_flubrd_symbol_set_p(0), do_moves(fix_d)
        elif a == 15: pass
        else:
            do_moves('d+')
            continue
        break


# Very similar methodology to the above; look for patterns of blocks with the
# right orientation, and then apply set-piece moves (depending on whether three
# or two pieces are out of position). In all cases, we loop until we have a
# solution because sometimes it takes more than one effort to get things
# correct (we could go for a more intelligent approach than this...)
def mellor_bottom_corner_orient ():
    base_colour = lookup_colour('d')
    fix_2 = 'f-f-r-d+r+f+d+f-u-f+d-f-r-d-r+u+f-f-'
    fix_3 = 'r-d-r+d-r-d-d-r+d-d-'
    while True:
        a = 0
        for i in xrange(4):
            rotated_flubrd_symbol_set_p(i)
            if lookup_colour('dfr') == base_colour:
                a += 1 << i
        if   a == 0:  rotated_flubrd_symbol_set_p(0), do_moves(fix_3)
        elif a == 8:  rotated_flubrd_symbol_set_p(0), do_moves(fix_3)
        elif a == 1:  rotated_flubrd_symbol_set_p(1), do_moves(fix_3)
        elif a == 2:  rotated_flubrd_symbol_set_p(2), do_moves(fix_3)
        elif a == 4:  rotated_flubrd_symbol_set_p(3), do_moves(fix_3)
        elif a == 6:  rotated_flubrd_symbol_set_p(0), do_moves(fix_2)
        elif a == 12: rotated_flubrd_symbol_set_p(1), do_moves(fix_2)
        elif a == 9:  rotated_flubrd_symbol_set_p(2), do_moves(fix_2)
        elif a == 3:  rotated_flubrd_symbol_set_p(3), do_moves(fix_2)
        elif a in (5, 10): do_moves(fix_2)
        else:   break


# Almost the same again; look at the pattern of correctly placed bottom edges,
# and apply set-piece moves to a logically rotated cube so that the pieces are
# in a certain pattern which we can solve. Again, it might take several efforts
# to get this right, so we loop until a solution is found (again we could do
# better than this if we tried!)
def mellor_bottom_edge_place ():
    fix = 'r+l-f+r-l+d-d-r+l-f+r-l+'
    while True:
        a = 0
        for i in xrange(4):
            rotated_flubrd_symbol_set_p(i)
            side_colour = lookup_colour('f')
            if  lookup_colour('fd') == side_colour or lookup_colour('df') == side_colour:
                a += 1 << i
        if   a == 0: do_moves(fix)
        elif a == 1: rotated_flubrd_symbol_set_p(0), do_moves(fix)
        elif a == 2: rotated_flubrd_symbol_set_p(1), do_moves(fix)
        elif a == 4: rotated_flubrd_symbol_set_p(2), do_moves(fix)
        elif a == 8: rotated_flubrd_symbol_set_p(3), do_moves(fix)
        else:   break


# A similar approach again, but looking for patterns and applying moves to get
# the bottom edges in the correct orientation.
def mellor_bottom_edge_orient ():
    base_colour = get_colour_symbolic(cube, 'd')
    fix_adjacent = 'f-f-r-r-r-u-d+b-b-u-u-d-d-f-u-f+u-u-d-d-b+b+u+d-r+u+r-r-f-f-'
    fix_opposite = 'r-r-l-l-r-u-d+b-b-u-u-d-d-f-u-u-f+u-u-d-d-b-b-u+d-r+u-u-r-r-l-l-'
    while True:
        a = 0
        for i in xrange(4):
            rotated_flubrd_symbol_set_p(i)
            if lookup_colour('df') == base_colour:
                a += (1 << i)
        if   a == 0:
            do_moves(fix_opposite)
            continue
        elif a == 3:  rotated_flubrd_symbol_set_p(2), do_moves(fix_adjacent)
        elif a == 6:  rotated_flubrd_symbol_set_p(3), do_moves(fix_adjacent)
        elif a == 12: rotated_flubrd_symbol_set_p(0), do_moves(fix_adjacent)
        elif a == 9:  rotated_flubrd_symbol_set_p(1), do_moves(fix_adjacent)
        elif a == 5:  rotated_flubrd_symbol_set_p(0), do_moves(fix_opposite)
        elif a == 10: rotated_flubrd_symbol_set_p(1), do_moves(fix_opposite)
        break


# Apply all the various stages for fixing a cube in order, but allow the user
# to specify when to stop.
def mellor_solve (stage):
    global cube
    cube = plugin_cube_state()
    if check_cube_structure(cube):
        stage_list = [mellor_top_edge_solve,
                    mellor_top_corner_solve,
                    mellor_middle_slice_solve,
                    mellor_bottom_corner_place,
                    mellor_bottom_corner_orient,
                    mellor_bottom_edge_place,
                    mellor_bottom_edge_orient]
        
        for i in xrange(stage):
            stage_list[i]()
        execute_move_buffer_p()


# Provide plenty of menu entries to give the user some control over the solving
# algorithm (he may want to perform part of the solution himself!)
scripts = [
        ('/'.join((_('Solvers'),_('Mellor (3x3)'))),                        lambda: mellor_solve(7)),
        ('/'.join((_('Solvers'),_('Mellor (3x3)'),_('Top edges'))),         lambda: mellor_solve(1)),
        ('/'.join((_('Solvers'),_('Mellor (3x3)'),_('Top slice'))),         lambda: mellor_solve(2)),
        ('/'.join((_('Solvers'),_('Mellor (3x3)'),_('Middle slice'))),      lambda: mellor_solve(3)),
        ('/'.join((_('Solvers'),_('Mellor (3x3)'),_('Bottom corner place'))),lambda: mellor_solve(4)),
        ('/'.join((_('Solvers'),_('Mellor (3x3)'),_('Bottom corner orient'))),lambda: mellor_solve(5)),
        ('/'.join((_('Solvers'),_('Mellor (3x3)'),_('Bottom edge place'))), lambda: mellor_solve(6)),
        ('/'.join((_('Solvers'),_('Mellor (3x3)'),_('Bottom edge orient'))),lambda: mellor_solve(7)),
    ]

