#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright (C) 2009  B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filename: main.c
# Original copyright and license:
#/*
#    GNUbik -- A 3 dimensional magic cube game.
#    Copyright (C) 1998, 2003--2004  John Darrington
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License,  or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/


import sys
from optparse import OptionParser

from . import config
from .confstore import confstore


def import_(pure_python):
    global ui, widget_set, selection
    
    if not pure_python:
        try:
            import drwBlock_c
            sys.modules['pybiklib.drwBlock'] = drwBlock_c
            sys.modules['pybiklib.cube'] = drwBlock_c
        except ImportError:
            print _('Module drwBlock_c not found, using fallback. Rendering will be slow.')

    import ui
    import widget_set
    import selection


def run(argv, frameQty, cube_dimension, solved):
    if frameQty is not None:
        ui.frameQty = frameQty
    
    confstore.init_notify()
    
    # Create the top level widget --- that is,  the main window which everything
    #   goes in
    main_window = widget_set.MainWindow(argv)

    # create the cube
    main_window.first_new_game(cube_dimension, solved)
    
    # initialise the selection mechanism
    selection.selection.finalize_init()
    
    main_window.start_main_loop()
    

# process the options we're interested in.
def app_opts(argv):
    parser = OptionParser(version='%prog '+config.VERSION)
    parser.add_option("-s", "--solved", action="store_true", default=False,
                  help=unicode(_("Start with the cube already solved"),'utf-8'))
    parser.add_option("--size", type="int", dest='cube_dimension',
                  help=unicode(_("Start with a N x N x N sized cube"),'utf-8'), metavar="N")
    parser.add_option("--animation", type="int", dest='frameQty',
                  help=unicode(_("Show N intermediate positions in animations"),'utf-8'), metavar="N")
    parser.add_option("--pure-python", action="store_true", default=False,
                  help=unicode(_("Use python module for rendering (very slow)"),'utf-8'))
    debug_level_names = 'func,msg,draw,keys,test'
    parser.add_option("--debug", type="string", dest='debug_level', default=-1,
                  help=unicode(_("Enable debug output, D: integer or [{0}]..."),'utf-8')
                                                        .format(debug_level_names),
                  metavar="D")
    
    options, args = parser.parse_args()
    if args:
        parser.error(_("incorrect number of arguments"))
    
    debug_level = options.debug_level
    debug_level_names = debug_level_names.upper().split(',')
    try:
        debug_level = int(debug_level)
        debug_level_max = sum([1<<i for i,n in enumerate(debug_level_names)])
        if not (-1 <= debug_level <= debug_level_max):
            parser.error(_("debug level out of range: {0}").format(debug_level))
    except ValueError:
        debug_level = debug_level.split(',')
        _debug_level_names = dict([(n,1<<i) for i,n in enumerate(debug_level_names)])
        debug_level = [(d.upper() if d.upper() in _debug_level_names else d) for d in debug_level]
        try:
            debug_level = sum([_debug_level_names[d] for d in debug_level])
        except KeyError as e:
            parser.error(_("unknown debug option: {0}").format(e.args[0]))
    options.debug_modes = {'DEBUG_LEVEL': debug_level}
    
    if debug_level < 0:
        debug_level = 0
    debug_modes = reversed(bin(debug_level)[2:].rjust(len(debug_level_names),'0'))
    debug_modes = [bool(int(c)) for c in debug_modes]
    debug_modes = zip(debug_level_names, debug_modes)
    options.debug_modes.update(dict(('DEBUG_'+n, v) for n,v in debug_modes))
    return options

