#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright (C) 2009  B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filename: quarternion.c
# Original copyright and license:
#/*
#
#    A little library to do quarternion arithmetic
#    Copyright (C) 2003  John Darrington
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License,  or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/


import math

#from .debug import *


class Quarternion:
    def __init__(self, w=0, x=0, y=0, z=0):
        self.w = w  # float
        self.x = x  # float
        self.y = y  # float
        self.z = z  # float


#def quarternion_set_to_unit (q):
#    q.w = 1
#    q.x = 0
#    q.y = 0
#    q.z = 0

#@debug_func
def quarternion_to_matrix (q):
#    for i in xrange(txfm.MATRIX_DIM):
#        for j in xrange(txfm.MATRIX_DIM):
#            M[i][j] = 99

    ww = q.w * q.w
    xx = q.x * q.x
    yy = q.y * q.y
    zz = q.z * q.z
    
    return [[ww + xx - yy - zz,       2*q.x*q.y - 2*q.w*q.z,      2*q.x*q.z + 2*q.w*q.y,  0.0],
            [2*q.x*q.y + 2*q.w * q.z, ww - xx + yy - zz,          2*q.y*q.z - 2*q.w*q.x,  0.0],
            [2*q.x*q.z - 2*q.w * q.y, 2*q.y*q.z + 2*q.w * q.x,    ww - xx - yy + zz,      0.0],
            [0.0,                     0.0,                        0.0,        ww + xx + yy + zz]]



def quarternion_from_rotation (q, u, theta):
    radians = theta * math.pi / 360.0

    s = math.sin(radians)

    q.w = math.cos(radians)
    q.x = u[0] * s
    q.y = u[1] * s
    q.z = u[2] * s


# Pre multiply q1 by q2
def quarternion_pre_mult (q1, q2):
    dot_product = q1.x*q2.x + q1.y*q2.y + q1.z*q2.z

    s1 = q1.w
    s2 = q2.w
    q1.w = q1.w * q2.w - dot_product

    cross_product = (q1.y*q2.z - q1.z*q2.y,
                     q1.z*q2.x - q1.x*q2.z,
                     q1.x*q2.y - q1.y*q2.x)

    q1.x = s1*q2.x + s2*q1.x + cross_product[0]
    q1.y = s1*q2.y + s2*q1.y + cross_product[1]
    q1.z = s1*q2.z + s2*q1.z + cross_product[2]


