#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright (C) 2009  B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filenames: select.c, select-common.c
# Original copyright and license from select.c:
#/*
#    A system to permit user selection of a block and rotation axis
#    of a magic cube.
#    Copyright (C) 1998,  2003  John Darrington
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License,  or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/
# Original copyright and license from select-common.c:
#/*
#  A system to permit user selection of a block and rotation axis
#  of a magic cube.
#  Copyright (C) 2003  John Darrington
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/


# This library provides  a means of picking a block using the mouse cursor.
# 
# Two mutually co-operative mechanisms are used in this library.  There is a
# timer callback,  which occurs at regular intervals.  There is also the mouse
# motion callback,  which occurs whenever the mouse cursor is moving.  If two
# consecutive timer callbacks occur,  without and intervening mouse motion callback,
# then the cursor is assumed to be stationary.
# 
# If a stationary mouse is detected,  the program goes on to determine which block
# in the cube (if any) the cursor is located upon.


import math

import gobject
from gtk import gdk

from OpenGL import GL as gl
from OpenGL import GLU as glu

#from .debug import *
import ui
import glarea
from . import glarea_common
import drwBlock
import cube
from . import txfm


class FacetSelection:
    def __init__(self, block=0, face=0, quadrant=0):
        self.block = block  # int
        self.face = face    # int
        self.quadrant = quadrant  # int
    def __str__(self):
        return 'FacetSelection(block=%s face=%s quadrant=%s)' % (self.block, self.face, self.quadrant)


BUFSIZE = 512

class Selection:
    def __init__(self):
        # is the selection mechanism necessary ? 
        self.needSelection = False
        # the time for which the mouse must stay still, for anything to happen.
        self.idle_threshold = 10
        # the minimum distance it must have moved.
        self.granularity = 1
        # function to be called when a new block is selected.
        self.selection_func_current = self.selection_func_none
        
        self.timerActive = False
        self.timer = 0
        
        self.enable_disable_selection_entered = 0
        
        self.mouse_x = -1
        self.mouse_y = -1
        self.motion = False
        self.stop_detected = False
        
        # Structure to hold copy of the last selection taken or None
        self.current_selection = None
    
    #@debug_func
    def finalize_init(self):
        '''Initialise the selection mechanism.'''
        
        self.selection_func_current = self.selection_func_original
        
        glarea.glwidget.connect("motion_notify_event", self._detect_motion, 0)
        
        self.timer = gobject.timeout_add(self.idle_threshold, self._unset_motion)
        self.timerActive = True
        
        # Add a handler to for all those occasions when we don't need the
        # selection mechanism going 
        
        glarea.glwidget.connect("enter-notify-event", self._enable_disable_selection, 0)
        glarea.glwidget.connect("leave-notify-event", self._enable_disable_selection, 0)
        glarea.glwidget.connect("visibility-notify-event", self._enable_disable_selection, 0)
        glarea.glwidget.connect("unmap-event", self._enable_disable_selection, 0)
        
    def disable(self):
        gobject.source_remove(self.timer)
        self.timerActive = False
        
    def enable(self):
        if not self.timerActive:
            self.timer = gobject.timeout_add(self.idle_threshold, self._unset_motion)
            self.timerActive = True
        self.needSelection = True
    
    def _enable_disable_selection(self, unused_w, event, unused_data):
        ''' When the window is not mapped, kill the selection mechanism.
        It wastes processor time'''
        # This is a kludge to work around a rather horrible bug;  for some
        # reason,  some  platforms emit a EnterNotify and LeaveNotify (in
        # that order) when animations occur.  This workaround makes sure
        # that the window is not `entered twice' 

        if event.type == gdk.ENTER_NOTIFY:
            self.enable_disable_selection_entered += 1
            self.needSelection = False
            if not self.timerActive:
                self.timer = gobject.timeout_add(self.idle_threshold, self._unset_motion)
                self.timerActive = True

        elif event.type == gdk.LEAVE_NOTIFY:
            self.update()
            self.enable_disable_selection_entered -= 1
            if self.enable_disable_selection_entered <= 0:
                self.needSelection = True
                gobject.source_remove(self.timer)
                self.timerActive = False

        return False

    # This callback occurs whenever the mouse is moving 
    def _detect_motion(self, unused_w, event, unused_user_data):
        if event.type != gdk.MOTION_NOTIFY:
            return False

        self.mouse_x = event.x
        self.mouse_y = event.y
        self.motion = True
        self.stop_detected = False
        return False

    def _unset_motion(self):
        '''This callback occurs at regular intervals. The period is determined by
        idle_threshold.  It checks to see if the mouse has moved,  since the last
        call of this function.
        Post-condition:  motion is False.'''
        if self.motion == False: # if not moved since last time 
            if not self.stop_detected:
                # in here,  things happen upon the mouse stopping 
                self.stop_detected = True
                self.update()
        
        self.motion = False
        return True

    def pick_polygons(self, x, y):
        '''Identify the block at screen co-ordinates x,y.
        
        This func determines all candidate blocks. That is, all blocks
        which orthogonally project to x,y. It then calls _choose_items,
        to determine which of them is closest to the screen.'''
        
        viewport = (0,)*4
        #selectBuf = [0]*BUFSIZE
        
        assert (self.granularity > 0)
        assert (glarea_common.fovy > 0)
        assert (glarea_common.cp_near > 0)
        assert (glarea_common.cp_far > 0)
        
        height = glarea.glwidget.allocation.height
        #debug('widget_height', height)
        
        gl.glSelectBuffer(BUFSIZE)
        gl.glRenderMode(gl.GL_SELECT)
        
        gl.glInitNames()
        gl.glPushName(0xFFFF)
        
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        
        viewport = gl.glGetIntegerv(gl.GL_VIEWPORT)
        glu.gluPickMatrix(x, height-y, self.granularity, self.granularity, viewport)
        glu.gluPerspective(glarea_common.fovy, 1, glarea_common.cp_near, glarea_common.cp_far)
        
        glarea_common.modelViewInit()
        drwBlock.pick_cube()
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPopMatrix()
        
        glarea.error_check("")
        #hits = gl.glRenderMode (gl.GL_RENDER)
        selectBuf = gl.glRenderMode(gl.GL_RENDER)
        try:
            hits = len(selectBuf)
        except TypeError:
            #XXX: ??? Under Virtualbox gl.glRenderMode(gl.GL_RENDER) returns len instead of buffer
            hits = selectBuf
            try:
                selectBuf = gl.glGetPointerv(gl.GL_SELECT)
            except:
                hits = 0
        #debug(hits, 'selectBuf', selectBuf and dir(selectBuf[0]))
        if hits == 0:
            #debug('hits == 0')
            glarea.error_check("Error Selecting")
            return None
        return self._choose_items(hits, selectBuf)
    
    @staticmethod
    def _choose_items(hits, buffer):
        '''Find out which of all the objects in buffer is the one with the
        lowest Z value.  ie. closer in the depth buffer'''
        
        assert hits == len(buffer)
        if not buffer:
            return None
        
        buf = min(buffer, key=lambda x:x.near)
        #buf = min(buffer, key=lambda x:x.far)
        #zvalue = buf.far
        closest = buf.names
        if len(closest) < 3:
            return None
        #SEL_BLOCK = 0
        #SEL_FACE = 1
        #SEL_QUAD = 2
        facet_selection = FacetSelection()
        facet_selection.block = closest[0]
        facet_selection.face = closest[1]
        facet_selection.quadrant = closest[2]
        #debug('facet_selection',facet_selection)
        return facet_selection
    
    def update(self):
        '''This func determines which block the mouse is pointing at, and
        calls the function "selection_func_current"'''
        
        #debug(self.current_selection)
        self.current_selection = self.pick_polygons(self.mouse_x, self.mouse_y)
        #debug(self.current_selection)
        self.selection_func_current()

    def is_item_selected(self):
        return self.current_selection is not None
        
    def selection_func_none(self):
        pass
    
    #@debug_func
    def selection_func_original(self):
        '''Called whenever a new set of polygons have been selected.'''
        if ui.is_animating():
            return
            
        if self.is_item_selected():
            turn_axis = ui.getTurnAxis(self.current_selection)
            ui.pending_movement.dir = ui.turnDir(turn_axis)
            ui.pending_movement.axis = ui.vector2axis(turn_axis)
            ui.pending_movement.slice = cube.get_pending_movement_slice(
                                                    self.current_selection.block,
                                                    ui.pending_movement.axis)
            
            # Here we take the orientation of the selected quadrant and multiply it
            # by the projection matrix.  The result gives us the angle (on the screen)
            # at which the mouse cursor needs to be drawn.
            proj = gl.glGetFloatv(gl.GL_PROJECTION_MATRIX)
            v = cube.get_quadrant_vector(self.current_selection.block,
                                        self.current_selection.face,
                                        self.current_selection.quadrant)
            v = txfm.transform(proj, v)
            glarea.cursorAngle = math.atan2(v[0], v[1]) * 180.0 / math.pi
        else:
            ui.pending_movement.dir = -1
            ui.pending_movement.axis = -1
        
        glarea.postSetCursor()

selection = Selection()


