#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright (C) 2009  B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filename: ui.c
# Original copyright and license:
#/*
#    User Interface functions for the Cube
#    Copyright (C) 1998,  2003  John Darrington
#                  2004  John Darrington,  Dale Mellor
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License,  or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/


import gtk.keysyms
import gobject

from .debug import *
from . import txfm
#TODO: move the code thst uses module cube to cube_form
import cube
import drwBlock
from . import glarea
from .selection import selection
from . import widget_set

from . import quarternion
from . import cube_form
from . import move_queue
from . import glarea_common
from .confstore import confstore


GNUBIK_Left = gtk.keysyms.Left
GNUBIK_Right = gtk.keysyms.Right
GNUBIK_Up  = gtk.keysyms.Up
GNUBIK_Down = gtk.keysyms.Down

# picture rate in ms. 40ms = 25 frames per sec
picture_rate = 40
# this is the number of frames drawn when a slice of the cube rotates 90 degrees
frameQty = confstore.frameQty

initial_cube_state = cube_form.CubeState()
current_cube_state = cube_form.CubeState()
# the one and only MoveQueue object
all_moves = move_queue.MoveQueue()

# The move taking place NOW
class _Current_Movement:
    def __init__(self):
        self.move_data = move_queue.MoveData()
        self.mark_before = False
        self.stop_requested = 0
        self.abort_requested = 0
        self.animation_in_progress = 0

current_movement = _Current_Movement()


# The move that will take place when the mouse is clicked
pending_movement = move_queue.MoveData(-1, -1, -1)


def is_animating():
    return current_movement.animation_in_progress


# Rotate cube about axis (screen relative)
def rotate_cube(axis, dir_):
    # how many degrees to turn the cube with each hit
    step = 2.0
    rot = quarternion.Quarternion()
    #vector v
    v = txfm.Vector()

    if dir_:
        step = -step

    if axis == 0:
        v[1] = v[2] = 0
        v[0] = 1
    elif axis == 1:
        v[0] = v[2] = 0
        v[1] = 1
    elif axis == 2:
        v[0] = v[1] = 0
        v[2] = 1

    quarternion.quarternion_from_rotation(rot, v, step)
    quarternion.quarternion_pre_mult(glarea_common.cube_view, rot)
    glarea.postRedisplay()


# orientate the whole cube with the arrow keys
def do_keysyms(keysym, shifted):
    if keysym == GNUBIK_Right:
        if shifted:
            dir_ = 0
            axis = 2
        else:
            dir_ = 1
            axis = 1
    elif keysym == GNUBIK_Left:
        if shifted:
            dir_ = 1
            axis = 2
        else:
            dir_ = 0
            axis = 1
    elif keysym == GNUBIK_Up:
        axis = 0
        dir_ = 0
    elif keysym == GNUBIK_Down:
        axis = 0
        dir_ = 1
    else:
        return

    rotate_cube(axis, dir_)
    return True


# return the turn direction,  based upon the turn axis vector
#@debug_func
def turnDir(vector):
    # This is a horrendous kludge.  It works only because we know that
    # vector is arithemtically very simple

    return vector[0] + vector[1] + vector[2]  >  0




# Convert a vector to a axis number.  Obviously this assumes the vector
#is orthogonal to the frame of reference ( another nasty kludge ).
def vector2axis(vector):
    return (0 if vector[0] != 0 else
            1 if vector[1] != 0 else
            2)




turn_axesX  = [1, 0, 0, 0]
turn_axes_X = [-1, 0, 0, 0]
turn_axesY  = [0, 1, 0, 0]
turn_axes_Y = [0, -1, 0, 0]
turn_axesZ  = [0, 0, 1, 0]
turn_axes_Z = [0, 0, -1, 0]
turn_axes = [
                [turn_axes_Y, turn_axesX,  turn_axesY,  turn_axes_X],
                [turn_axesY,  turn_axesX,  turn_axes_Y, turn_axes_X],
                [turn_axesZ,  turn_axesX,  turn_axes_Z, turn_axes_X],
                [turn_axes_Z, turn_axesX,  turn_axesZ,  turn_axes_X],
                [turn_axes_Y, turn_axes_Z, turn_axesY,  turn_axesZ],
                [turn_axesY,  turn_axes_Z, turn_axes_Y, turn_axesZ],
            ]

# Determine the axis about which to rotate the slice,  from the objects
# selected by the cursor position
#@debug_func
def getTurnAxis(items):
    # Each edge (quadrant) on a block represents a different axis

    # Fetch the selected block's transformation from its original
    # orientation
    t = cube.get_block_transform(items.block)

    # Select the untransformed vector for the selected edge
    vector = [turn_axes[items.face][items.quadrant][i] for i in xrange(4)]

    # transform it,  so that we go the right way
    vector = txfm.transform(t, vector)
    
    return vector


# handle mouse clicks
def mouse(button, application):
    # Don't let a user make a move,  whilst one is already in progress,
    # otherwise the cube falls to bits.
    if current_movement.animation_in_progress:
        return

    if button == 1:
        # Make a move
        if selection.is_item_selected():
            request_rotation(pending_movement, application)
        
        glarea.postRedisplay()


### Funktions to control the cube for use in callbacks and scripts

#@debug_func
def request_stop():
    '''Stop when the current animation completes'''
    current_movement.stop_requested = 1

def request_abort():
    '''Stop as soon as possible.'''
    current_movement.abort_requested = 1

#@debug_func
def request_back(application):
    '''One step back in the sequence of moves'''
    assert all_moves
    if not current_movement.animation_in_progress and not all_moves.at_start():
        all_moves.retard()
        current_movement.stop_requested = 1
        next_move = all_moves.current().copy()
        next_move.dir = not next_move.dir
        current_movement.move_data = next_move
        current_movement.mark_before = next_move.mark_after
        animate_rotation(application)

# Internal functionality for the following two functions.
#@debug_func
def _request_play(one_move, application):
    assert all_moves
    if not current_movement.animation_in_progress and all_moves.current():
        current_movement.stop_requested = one_move
        current_movement.move_data = all_moves.current().copy()
        current_movement.mark_before = (True if all_moves.at_start() else 
                                        all_moves.prev().mark_after)
        all_moves.advance()
        animate_rotation(application)
        return True
    return False

#@debug_func
def request_play(application):
    '''Play the current sequence of moves to the end'''
    _request_play(0, application)

#@debug_func
def request_next(application):
    '''One step forward in the sequence of moves'''
    _request_play(1, application)

#@debug_func
def request_rotation(data, application):
    '''Make one new move.
    
    The usual rotation request,  called when user uses mouse to rotate the cube by
    hand. The move is put into the `current' place,  and all the moves in 
    all_moves afterwards are zapped.'''
    assert all_moves
    all_moves.push_current(data)
    application.update_statusbar()
    request_play(application)

#@debug_func
def request_delayed_rotation(maxis, mslice, mdir):
    '''Apppend one step to the sequence of moves.
    
    The rotation request called from scripts. In this case the moves are
    appended to all_moves,  so the script can race ahead and work out the
    moves,  leaving the graphics to catch up later.'''
    assert all_moves
    
    data = move_queue.MoveData(mslice, mdir, maxis)
    all_moves.push(data)

#@debug_func
def request_truncate_move_queue():
    '''Discard moves from current position to the end in the sequence of moves.
    
    Scripts calls this function on startup,  and then the above function as each
    move is computed. This way,  the moves computed are appended to all_moves
    from the current point forwards (and forward data that was on the queue is
    discarded).'''
    assert all_moves
    all_moves.truncate()

#@debug_func
def request_mark_move_queue(application=None, mark=True):
    '''Set a marker at the current position in the sequence of moves'''
    assert all_moves
    all_moves.mark_current(mark)
    # if called from scripts update_ui is called later
    if application:
        update_ui(application)

#@debug_func
def request_queue_rewind(application, silent=False, to_mark=True):
    '''Go to first position in the sequence of moves.
    
    All the moves take place on the cube,  not via the animation
    routines,  so the effect is to unwind the queue instantaneously.'''
    assert all_moves

    if all_moves.at_start():
        return
    move_data = all_moves.prev()
    while True:
        all_moves.retard()
        cube.rotate_back(move_data)
        if all_moves.at_start():
            break
        move_data = all_moves.prev()
        if to_mark and move_data.mark_after:
            break
    if not silent:
        #TODO: postRedisplay updates internal cube state in at idle time
        # maybe that is the reason for the following bug
        # internal state should applied before gobject.idle_add
        # this is also for request_fast_forward
        glarea.postRedisplay()
        #update_ui(application)
        #FIXME: sometimes 'rotated center' is shown by mistake instead of 'solved'
        gobject.idle_add(update_ui, application)

#@debug_func
def request_fast_forward(application=None, to_pos=-1):
    '''Go to end of the sequence of moves.
    
    All the moves take place on the cube,  not via the animation
    routines,  so the effect is to show the result instantaneously.'''
    assert all_moves

    move_data = all_moves.current()
    while move_data:
        if to_pos == all_moves.current_place:
            break
        current_cube_state.rotate_slice(move_data)
        all_moves.advance()
        if to_pos < 0 and move_data.mark_after:
            break
        move_data = all_moves.current()

    glarea.postRedisplay()
    if application:
        #update_ui(application)
        gobject.idle_add(update_ui, application)

@debug_func
def request_parse_formula(application, code, pos):
    request_queue_rewind(application, silent=True, to_mark=False)
    #debug('textpos', pos)
    pos = all_moves.parse(code, pos, current_cube_state.dimension)
    #debug('queuepos', pos)
    request_fast_forward(application, to_pos=pos)


### Helper functions

#@debug_func
def animate_rotation(application):
    assert type(application)==widget_set.MainWindow, application
    
    current_movement.animation_in_progress = 1
    drwBlock.start_animate(current_movement.move_data)
    application.playbarstate.playing(current_movement.mark_before)
    gobject.timeout_add(picture_rate, on_animate, application)

# a timeout  calls this func,  to animate the cube
#@debug_func
def on_animate(application):
    # how many degrees motion per frame
    increment = 90.0 / (frameQty + 1)

    # decrement the current angle
    unfinished = drwBlock.step_animate(increment)

    # and redraw it
    glarea.postRedisplay ()

    if unfinished and not current_movement.abort_requested:
        # call this timeout again instead of return True; picture_rate may changed
        gobject.timeout_add(picture_rate, on_animate, application)
    else:
        # we have finished the animation sequence now
        drwBlock.end_animate(current_movement.move_data)

        current_movement.animation_in_progress = 0
        
        selection.update()
        
        if current_movement.abort_requested:
            current_movement.abort_requested = 0

        # If there are more animations in the queue,  go again unless a stop has
        # been requested.

        if all_moves.current() and not current_movement.stop_requested:
            application.update_statusbar()
            move_data = all_moves.current().copy()
            current_movement.move_data = move_data
            current_movement.mark_before = (True if all_moves.at_start() else 
                                            all_moves.prev().mark_after)
            all_moves.advance()
            animate_rotation(application)
        else:
            current_movement.stop_requested = 0
            update_ui(application)
        
    return False

# in cube.py        NOT_SOLVED = 0,  SOLVED = 1,  HALF_SOLVED = 2
#map_cube_status = [_('not solved'), _('solved'), _('rotated center')]
#TODO: HALF_SOLVED is broken in some ways
#       * see comment in request_queue_rewind
#       * should only shown if asymetric pattern or image
pass
map_cube_status = [_('not solved'), _('solved'), _('solved')]

#@debug_func
def update_ui(application):
    assert type(application)==widget_set.MainWindow, application
    
    if all_moves.at_start():
        if all_moves.at_end():
            application.playbarstate.empty()
        else:
            application.playbarstate.first()
    else:
        if all_moves.at_end():
            application.playbarstate.last(all_moves.tail.mark_after)
        else:
            application.playbarstate.mid(all_moves.prev().mark_after)
        
    code, pos, markup = all_moves.format(current_cube_state.dimension)
    application.set_entry_formula(code, pos, markup)
    status = cube.cube_status_check()
    application.update_statusbar(map_cube_status[status])
    

@debug_func
def update_ui_first(application):
    assert type(application)==widget_set.MainWindow, application
    
    if all_moves.at_start():
        if all_moves.at_end():
            application.playbarstate.empty()
        else:
            application.playbarstate.first()
    else:
        if all_moves.at_end():
            application.playbarstate.last(all_moves.tail.mark_after)
        else:
            application.playbarstate.mid(all_moves.prev().mark_after)
        
    code, pos, markup = all_moves.format(current_cube_state.dimension)
    application.set_entry_formula(code, pos, markup)
    status = cube.cube_status_check()
    application.update_statusbar_first(map_cube_status[status])
    

#if DEBUG_KEYS:
#    # Manually make a move on the cube
#    def on_manual_move(unused_w, data, application):
#        # Just loose the first argument and call the next function
#        request_rotation(data, application)
#        glarea.postRedisplay()

