#!/usr/bin/env python

from glob import glob
import os
import sys
import subprocess

import distutils

from distutils.core import setup
from distutils.cmd import Command
from distutils.extension import Extension

from distutils.errors import DistutilsFileError, DistutilsOptionError

from distutils.dir_util import remove_tree
from distutils.log import warn, info, error, fatal

import distutils.command.build
import distutils.command.clean
import distutils.command.build_ext
from distutils.command.install_data import install_data

#from Pyrex.Distutils.extension import Extension
#from Pyrex.Distutils import build_ext
#from distutils.extension import Extension
#from Cython.Distutils import build_ext

from pybiklib import version


############ This code is taken from python-distutils-extra
class build_extra(distutils.command.build.build):
    """Adds the extra commands to the build target. This class should be used
       with the core distutils"""
    def __init__(self, dist):
        distutils.command.build.build.__init__(self, dist)

        self.user_options.extend([("i18n", None, "use the localsation"),
                                  ("icons", None, "use icons"),
                                  ("help", None, "use help system")])
    def initialize_options(self):
        distutils.command.build.build.initialize_options(self)
        self.i18n = False
        self.icons = False
        self.help = False

    def finalize_options(self):
        def has_help(command):
            return self.help == "True" or \
                   (self.distribution.cmdclass.has_key("build_help") and not \
                    self.help == "False")
        def has_icons(command):
            return self.icons == "True" or \
                   (self.distribution.cmdclass.has_key("build_icons") and not \
                    self.help == "False")
        def has_i18n(command):
            return self.i18n == "True" or \
                   (self.distribution.cmdclass.has_key("build_i18n") and not \
                    self.i18n == "False")
        distutils.command.build.build.finalize_options(self)
        self.sub_commands.append(("build_i18n",  has_i18n))
        self.sub_commands.append(("build_icons", has_icons))
        self.sub_commands.append(("build_help",  has_help))

class build(build_extra):
    """Adds the extra commands to the build target. This class should be
       used with setuptools."""
    def finalize_options(self):
        distutils.command.build.build.finalize_options(self)
        self.sub_commands.append(("build_i18n",  lambda cmd: self.i18n == "True"))
        self.sub_commands.append(("build_icons", lambda cmd: self.icons == "True"))
        self.sub_commands.append(("build_help",  lambda cmd: self.help == "True"))

class build_i18n(distutils.cmd.Command):

    description = "integrate the gettext framework"

    user_options = [
        ('desktop-files=',  None, '.desktop.in files that should be merged'),
        ('xml-files=',      None, '.xml.in files that should be merged'),
        ('schemas-files=',  None, '.schemas.in files that should be merged'),
        ('key-files=',      None, '.key.in files that should be merged'),
        ('domain=',         'd',  'gettext domain'),
        ('merge-po',        'm',  'merge po files against template'),
        ('po-dir=',         'p',  'directory that holds the i18n files'),
        ('bug-contact=',    None, 'contact address for msgid bugs')]

    boolean_options = ['merge-po']

    def initialize_options(self):
        self.desktop_files = []
        self.xml_files = []
        self.key_files = []
        self.schemas_files = []
        self.domain = None
        self.merge_po = False
        self.bug_contact = None
        self.po_dir = None

    def finalize_options(self):
        if self.domain is None:
            self.domain = self.distribution.metadata.name
        if self.po_dir is None:
            self.po_dir = "po"

    def run(self):
        """
        Update the language files, generate mo files and add them
        to the to be installed files
        """
        data_files = self.distribution.data_files

        if self.bug_contact is not None:
            os.environ["XGETTEXT_ARGS"] = "--msgid-bugs-address=%s " % \
                                          self.bug_contact

        # Print a warning if there is a Makefile that would overwrite our
        # values
        if os.path.exists("%s/Makefile" % self.po_dir):
            self.announce("""
WARNING: Intltool will use the values specified from the
         existing po/Makefile in favor of the vaules
         from setup.cfg.
         Remove the Makefile to avoid problems.""")

        # Update po(t) files and print a report
        # We have to change the working dir to the po dir for intltool
        cmd = ["intltool-update", (self.merge_po and "-r" or "-p"), "-g", self.domain]
        wd = os.getcwd()
        os.chdir(self.po_dir)
        self.spawn(cmd)
        os.chdir(wd)

        for po_file in glob("%s/*.po" % self.po_dir):
            lang = os.path.basename(po_file[:-3])
            mo_dir =  os.path.join("build", "mo", lang, "LC_MESSAGES")
            mo_file = os.path.join(mo_dir, "%s.mo" % self.domain)
            if not os.path.exists(mo_dir):
                os.makedirs(mo_dir)
            cmd = ["msgfmt", po_file, "-o", mo_file]
            self.spawn(cmd)

            targetpath = os.path.join("share/locale", lang, "LC_MESSAGES")
            data_files.append((targetpath, (mo_file,)))

        # merge .in with translation
        for (option, switch) in ((self.xml_files, "-x"),
                                 (self.desktop_files, "-d"),
                                 (self.schemas_files, "-s"),
                                 (self.key_files, "-k"),):
            try:
                file_set = eval(option)
            except:
                continue
            for (target, files) in file_set:
                build_target = os.path.join("build", target)
                if not os.path.exists(build_target): 
                    os.makedirs(build_target)
                files_merged = []
                for f in files:
                    if f.endswith(".in"):
                        file_merged = os.path.basename(f[:-3])
                    else:
                        file_merged = os.path.basename(f)
                    file_merged = os.path.join(build_target, file_merged)
                    cmd = ["intltool-merge", switch, self.po_dir, f, 
                           file_merged]
                    self.spawn(cmd)
                    files_merged.append(file_merged)
                data_files.append((target, files_merged))

class clean_i18n(distutils.command.clean.clean):
    description = "clean up files generated by build_i18n"

    def run(self):
        # clean build/mo
        mo_dir =  os.path.join("build", "mo")
        if os.path.isdir(mo_dir):
            remove_tree('build/mo')

        # clean built i18n files
        for setname in ('xml_files', 'desktop_files', 'schemas_files', 'key_files'):
            file_set = eval(self.distribution.get_option_dict('build_i18n').get(setname, (None, '[]'))[1])
            for (target, files) in file_set:
                build_target = os.path.join("build", target)
                for f in files:
                    if f.endswith(".in"):
                        file_merged = os.path.basename(f[:-3])
                    else:
                        file_merged = os.path.basename(f)
                    file_merged = os.path.join(build_target, file_merged)
                    if os.path.exists(file_merged):
                        os.unlink(file_merged)
                if os.path.exists(build_target):
                    os.removedirs(build_target)

        distutils.command.clean.clean.run(self)
#####################


def py2pyx(infile, outfile):
    info('py2pyx: %s --> %s' % (infile, outfile))
    os.system('python tools/py2pyx.py %s %s' % (infile, outfile))
def pyrex(infile, outfile):
    info('pyrex: %s --> %s' % (infile, outfile))
    os.system("pyrexc %s" % infile)
def cython(infile, outfile):
    info('cython: %s --> %s' % (infile, outfile))
    os.system("cython -a {0} 2>{0}.lis; grep '^/' {0}.lis".format(infile))


class build_ext (distutils.command.build_ext.build_ext):
    user_options = distutils.command.build_ext.build_ext.user_options + [
                            ('cython',None,'Use Cython to compile pyx-files'),
                            ('pyrex',None,'Use pyrex to compile pyx-files')]
    def __init__(self, *args):
        self.cython = None
        self.pyrex = None
        distutils.command.build_ext.build_ext.__init__(self, *args)
        
    def finalize_options(self):
        distutils.command.build_ext.build_ext.finalize_options(self)
        if self.cython is None:
            if self.pyrex is None:
                # Default Value
                self.cython = True
            else:
                self.cython = not self.pyrex
        else:
            if self.pyrex is None:
                self.cython = bool(self.cython)
            else:
                raise DistutilsOptionError('Option pyrex and cython conflicts')
        del self.pyrex
        
    def swig_sources(self, sources, extension):
        in_files = sources + extension.depends
        pyx_files = []
        pyx_files_dep = []
        for in_file in in_files:
            out_files = pyx_files if in_file in sources else pyx_files_dep
            if in_file.endswith('.py'):
                out_file = os.path.join(self.build_temp, in_file[:-3]+'_c.pyx')
                self.mkpath(os.path.dirname(out_file))
                self.make_file(in_files, out_file, py2pyx, (in_file, out_file))
            elif in_file.endswith(('.pxi','.pxd')):
                out_file = os.path.join(self.build_temp, in_file)
                self.mkpath(os.path.dirname(out_file))
                self.copy_file(in_file, out_file)
            else:
                out_file = in_file
            out_files.append(out_file)
        out_files = []
        for in_file in pyx_files:
            if in_file.endswith('.pyx'):
                out_file = in_file[:-4]+'.c'
                self.make_file(pyx_files+pyx_files_dep, out_file,
                                cython if self.cython else pyrex, (in_file, out_file))
                out_files.append(out_file)
            else:
                out_files.append(in_file)
        return distutils.command.build_ext.build_ext.swig_sources(self, out_files, extension)


class uninstall(Command):
    description = "Attempt an uninstall from an install --record file"

    user_options = [('manifest=', None, 'Installation record filename')]

    def initialize_options(self):
        self.manifest = None

    def finalize_options(self):
        pass

    def get_command_name(self):
        return 'uninstall'

    def run(self):
        self.ensure_filename('manifest')
        try:
            try:
                mfile = open(self.manifest)
                files = [f.strip() for f in mfile]
            except IOError, e:
                raise DistutilsFileError("unable to open install manifest: %s", str(e))
        finally:
            mfile.close()

        for f in files:
            if os.path.isfile(f) or os.path.islink(f):
                info("removing %s" % repr(f))
                if not self.dry_run:
                    try:
                        os.unlink(f)
                    except OSError, e:
                        warn("could not delete: %s" % repr(f))
            elif not os.path.isdir(f):
                info("skipping %s" % repr(f))

        dirs = set()
        for f in reversed(sorted(files)):
            dir_ = os.path.dirname(f)
            if dir_ not in dirs and os.path.isdir(dir) and len(os.listdir(dir_)) == 0:
                dirs.add(dir_)
                # Only nuke empty Python library directories, else we could destroy
                # e.g. locale directories we're the only app with a .mo installed for.
                if dir_.find("site-packages/") > 0:
                    info("removing %s" % repr(dir_))
                    if not self.dry_run:
                        try:
                            os.rmdir(dir_)
                        except OSError, e:
                            warn("could not remove directory: %s" % str(e))
                else:
                    info("skipping empty directory %s" % repr(dir_))



setup(  name=version.PACKAGE,
        version=version.VERSION,
        description='Pybik - the magic cube',
        author=version.AUTHORS[0],
        author_email=version.CONTACT,
        url=version.WEBSITE,
        license=version.LICENSE_NAME,
        scripts=['pybik'],
        data_files=[
                        ('share/applications', ['data/pybik.desktop']),
                        ('share/pixmaps', ['data/pybik.png']),
                    ],
        ext_modules=[ 
                        Extension("pybiklib/drwBlock_c",
                                    ["pybiklib/drwBlock.py"],
                                    depends=["pybiklib/cube.py",
                                            #"pybiklib/cube_c_.pxd",
                                            'pybiklib/opengl_gl.pxi',
                                            ],
                                    libraries=['GL']),
                    ],
        packages=['pybiklib','pybiklib/data'],
        package_dir={'pybiklib': 'pybiklib', 'pybiklib/data': 'data', },
        package_data={'pybiklib/data': [
                        '*.py',
                        'pybik.ui',
                        'pybik.png',
                    ]},
        cmdclass={  #'build': BuildData,
                    'build': build_extra,
                    'build_ext': build_ext,
                    'install_data': install_data,
                    'uninstall': uninstall,
                    'build_i18n': build_i18n,
                    'clean_i18n': clean_i18n,
                },
     )

